"""
You can use Propan MQBrokers without PropanApp
Just start and stop them whenever you want
"""
import falcon
import falcon.asgi
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")


@broker.handle("test")
async def base_handler(body):
    print(body)


class ThingsResource:
    async def on_get(self, req, resp):
        resp.status = falcon.HTTP_200
        resp.content_type = falcon.MEDIA_TEXT
        resp.text = (
            "\nTwo things awe me most, the starry sky "
            "above me and the moral law within me.\n"
            "\n"
            "    ~ Immanuel Kant\n\n"
        )


class PropanMiddleware:
    async def process_startup(self, scope, event):
        await broker.start()

    async def process_shutdown(self, scope, event):
        await broker.close()


app = falcon.asgi.App()
app.add_middleware(PropanMiddleware())
app.add_route("/things", ThingsResource())



import json
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Union
from uuid import uuid4

import nats
from nats.aio.client import Client
from nats.aio.msg import Msg
from propan.brokers.model import BrokerUsecase, ContentTypes
from propan.brokers.nats.schemas import Handler
from propan.brokers.push_back_watcher import BaseWatcher
from propan.types import DecoratedCallable
from propan.utils.context.main import log_context


class NatsBroker(BrokerUsecase):
    handlers: List[Handler]
    _connection: Optional[Client]

    __max_queue_len: int
    __max_subject_len: int

    def __init__(self, *args: Any, log_fmt: Optional[str] = None, **kwargs: Any):
        super().__init__(*args, log_fmt=log_fmt, **kwargs)

        self._connection = None

        self.__max_queue_len = 0
        self.__max_subject_len = 4

    async def _connect(self, *args: Any, **kwargs: Any) -> Client:
        return await nats.connect(*args, **kwargs)

    def handle(
        self, subject: str, queue: str = "", retry: Union[bool, int] = False
    ) -> Callable[[DecoratedCallable], None]:
        i = len(subject)
        if i > self.__max_subject_len:
            self.__max_subject_len = i

        i = len(queue)
        if i > self.__max_queue_len:
            self.__max_queue_len = i

        parent = super()

        def wrapper(func: DecoratedCallable) -> None:
            for handler in self.handlers:
                if handler.subject == subject and handler.queue == queue:
                    raise ValueError(
                        f"`{func.__name__}` uses already "
                        f"using `{subject}` subject with "
                        f"`{queue}` queue"
                    )

            func = parent.handle(func, retry, queue=queue, subject=subject)
            handler = Handler(callback=func, subject=subject, queue=queue)
            self.handlers.append(handler)

        return wrapper

    async def start(self) -> None:
        await super().start()

        for handler in self.handlers:
            func = handler.callback

            if self.logger:
                self._get_log_context(None, handler.subject, handler.queue)
                self.logger.info(f"`{func.__name__}` waiting for messages")

            sub = await self._connection.subscribe(handler.subject, cb=func)
            handler.subscription = sub

    async def publish_message(
        self, message: Union[str, Dict[str, Any]], subject: str, **publish_args: Any
    ) -> None:
        if self._connection is None:
            raise ValueError("NatsConnection not started yet")

        if isinstance(message, dict):
            message = json.dumps(message)
            headers = {
                **publish_args.pop("headers", {}),
                "content-type": ContentTypes.json.value,
            }
        else:
            headers = {
                **publish_args.pop("headers", {}),
                "content-type": ContentTypes.text.value,
            }

        return await self._connection.publish(
            subject, message.encode(), headers=headers, **publish_args
        )

    async def close(self) -> None:
        for h in self.handlers:
            await h.subscription.unsubscribe()

        if self._connection:
            await self._connection.drain()

    def _get_log_context(
        self, message: Optional[Msg], subject: str, queue: str = "", **kwargs
    ) -> Dict[str, Any]:
        if message is not None:
            message_id = message.reply or uuid4().hex
            message.message_id = message_id

        context = {
            "subject": subject,
            "queue": queue,
            "message_id": message.message_id[:10] if message else "",
        }

        log_context.set(context)
        return context

    @property
    def fmt(self) -> str:
        return self._fmt or (
            "%(asctime)s %(levelname)s - "
            f"%(subject)-{self.__max_subject_len}s | "
            + (f"%(queue)-{self.__max_queue_len}s | " if self.__max_queue_len else "")
            + "%(message_id)-10s "
            "- %(message)s"
        )

    @staticmethod
    async def _decode_message(message: Msg) -> Union[str, dict, bytes]:
        body = message.data
        if message.header:
            content_type = message.header.get("content-type", "")
            if ContentTypes.json.value in content_type:
                body = json.loads(body.decode())
            elif ContentTypes.text.value in content_type:
                body = body.decode()
        return body

    @staticmethod
    def _process_message(
        func: DecoratedCallable, watcher: Optional[BaseWatcher] = None
    ) -> Callable[[Msg], Any]:
        @wraps(func)
        async def wrapper(message: Msg):
            return await func(message)

        return wrapper



import asyncio
import logging
from ctypes import Union
from typing import Dict, List, Optional

from anyio import create_memory_object_stream, create_task_group
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from propan.brokers.model.broker_usecase import BrokerUsecase
from propan.cli.supervisors.utils import set_exit
from propan.log import logger
from propan.types import DecoratedCallableNone
from propan.utils import apply_types, context
from propan.utils.functions import to_async


class PropanApp:
    _on_startup_calling: List[DecoratedCallableNone]
    _on_shutdown_calling: List[DecoratedCallableNone]

    _stop_stream: Optional[MemoryObjectSendStream]
    _receive_stream: Optional[MemoryObjectReceiveStream]

    def __init__(
        self,
        broker: Optional[BrokerUsecase] = None,
        logger: Optional[logging.Logger] = logger,
    ):
        self.broker = broker
        self.logger = logger
        self.context = context
        context.set_context("app", self)

        self._on_startup_calling = []
        self._on_shutdown_calling = []
        self._stop_stream = None
        self._receive_stream = None
        self._command_line_options: Dict[str, Union[str, bool]] = {}

    def set_broker(self, broker: BrokerUsecase) -> None:
        self.broker = broker

    def on_startup(self, func: DecoratedCallableNone) -> DecoratedCallableNone:
        self._on_startup_calling.append(apply_types(to_async(func)))
        return func

    def on_shutdown(self, func: DecoratedCallableNone) -> DecoratedCallableNone:
        self._on_shutdown_calling.append(apply_types(to_async(func)))
        return func

    async def run(self, log_level: int = logging.INFO) -> None:
        self._init_async_cycle()
        async with create_task_group() as tg:
            set_exit(lambda *_: tg.start_soon(self._stop_stream.send, True))
            tg.start_soon(self._stop, log_level)
            tg.start_soon(self._start, log_level)

    def _init_async_cycle(self):
        self.loop = asyncio.get_event_loop()
        if not self._stop_stream and not self._receive_stream:
            self._stop_stream, self._receive_stream = create_memory_object_stream(1)

    async def _start(self, log_level: int = logging.INFO) -> None:
        self._log(log_level, "Propan app starting...")
        await self._startup()
        self._log(log_level, "Propan app started successfully! To exit press CTRL+C")

    async def _stop(self, log_level: int = logging.INFO) -> None:
        await self._receive_stream.receive()
        self._log(log_level, "Propan app shutting down...")
        await self._shutdown()
        self._log(log_level, "Propan app shut down gracefully.")

    def _log(self, level: int, message: str) -> None:
        if self.logger is not None:
            self.logger.log(level, message)

    async def _startup(self) -> None:
        for func in self._on_startup_calling:
            await func(**self._command_line_options)

        if self.broker is not None:
            await self.broker.start()

    async def _shutdown(self) -> None:
        if (
            self.broker is not None
            and getattr(self.broker, "_connection", False) is not False
        ):
            await self.broker.close()

        for func in self._on_shutdown_calling:
            await func()



"""
Alias is able to provide access to specific attributes of dependency
"""
from propan import Context, PropanApp
from propan.annotations import ContextRepo
from propan.brokers.rabbit import RabbitBroker
from pydantic import BaseSettings, Field

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)

KEY = "my-secret-key"


class Settings(BaseSettings):
    key: str = Field("")


@app.on_startup
async def setup(context: ContextRepo):
    settings = Settings(key=KEY)
    context.set_context("settings", settings)


@app.on_startup
def setup_later(key: str = Context("settings.key")):
    assert key is KEY



"""
@apply_types decorator allows pass context dependencies
to all functions with the same context through the functions
calling stack.
"""
from propan import PropanApp, apply_types
from propan.annotations import Logger, RabbitMessage
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


@broker.handle("test")
async def base_handler(body: dict):
    await nested_funcion()


@apply_types
async def nested_funcion(logger: Logger, message: RabbitMessage):
    logger.info(f"Message `{message}` processing")



"""
You can use Propan MQBrokers without PropanApp
Just start and stop them whenever you want
"""
import asyncio

import tornado.web
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")


@broker.handle("test")
async def base_handler(body):
    print(body)


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.write("Hello, world")


def make_app():
    return tornado.web.Application(
        [
            (r"/", MainHandler),
        ]
    )


async def main():
    app = make_app()
    app.listen(8888)

    await broker.start()
    try:
        await asyncio.Event().wait()
    finally:
        await broker.close()


if __name__ == "__main__":
    asyncio.run(main())



# Imports to use at __all__
from propan.cli.app import *  # noqa: F403
from propan.log import *  # noqa: F403
from propan.utils import *  # noqa: F403

__all__ = (  # noqa: F405
    # app
    "PropanApp",
    # log
    "logger",
    "access_logger",
    # utils
    ## types
    "apply_types",
    ## context
    "context",
    "Context",
    "ContextRepo" "Alias",
    "Depends",
)



import logging
import logging.config
from functools import partial
from typing import Any, Dict, Type

from propan.log.formatter import AccessFormatter, DefaultFormatter


def configure_formatter(
    formatter: Type[logging.Formatter], *args: Any, **kwargs: Any
) -> logging.Formatter:
    return formatter(*args, **kwargs)


LOGGING_CONFIG: Dict[str, Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": partial(configure_formatter, DefaultFormatter),
            "fmt": "%(asctime)s %(levelname)s - %(message)s",
            "use_colors": True,
        },
        "access": {
            "()": partial(configure_formatter, AccessFormatter),
            "fmt": "%(asctime)s %(levelname)s - %(message)s",
            "use_colors": True,
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "propan": {"handlers": ["default"], "level": "INFO", "propagate": False},
        "propan.error": {"level": "INFO"},
        "propan.access": {"handlers": ["access"], "level": "INFO", "propagate": False},
    },
}

logging.config.dictConfig(LOGGING_CONFIG)


logger = logging.getLogger("propan")
access_logger = logging.getLogger("propan.access")



import logging

from propan.cli.app import PropanApp
from propan.utils.context import Context as ContextField
from propan.utils.context import ContextRepo as CR
from typing_extensions import Annotated

Logger = Annotated[logging.Logger, ContextField("logger")]
App = Annotated[PropanApp, ContextField("app")]
ContextRepo = Annotated[CR, ContextField("context")]

RabbitBroker = NatsBroker = RabbitMessage = NatsMessage = None

try:
    import aio_pika
    from propan.brokers.rabbit import RabbitBroker as RB

    RabbitBroker = Annotated[RB, ContextField("broker")]
    RabbitMessage = Annotated[aio_pika.IncomingMessage, ContextField("message")]
except Exception:
    pass

try:
    from nats.aio.msg import Msg
    from propan.brokers.nats import NatsBroker as NB

    NatsBroker = Annotated[NB, ContextField("broker")]
    NatsMessage = Annotated[Msg, ContextField("message")]
except Exception:
    pass

assert any(
    (
        all((RabbitBroker, RabbitMessage)),
        all((NatsBroker, NatsMessage)),
    )
), "You should specify using broker!"



from functools import reduce
from typing import Dict, List, Tuple, Union


def parse_cli_args(*args: str) -> Tuple[str, Dict[str, Union[bool, str]]]:
    extra_kwargs: Dict[str, Union[bool, str]] = {}

    k: str = ""
    v: Union[bool, str, None] = None

    field_args: List[str] = []
    app = ""
    for item in reduce(lambda acc, x: acc + x.split("="), args, []) + ["-"]:
        if ":" in item:
            app = item

        else:
            if "-" in item:
                if k:
                    k = k.strip().lstrip("-").replace("-", "_")

                    if len(field_args) == 0:
                        v = not k.startswith("no_")
                    elif len(field_args) == 1:
                        v = field_args[0]
                    else:
                        v = field_args

                    extra_kwargs[remove_prefix(k, "no_")] = v
                    field_args = []

                k = item

            else:
                field_args.append(item)

    return app, extra_kwargs


def remove_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text



import logging
from ssl import SSLContext
from typing import Any, Dict, List, Optional, Union

import aio_pika
import aiormq
from pamqp.common import FieldTable
from propan.brokers.model import BrokerUsecase
from propan.brokers.push_back_watcher import BaseWatcher
from propan.brokers.rabbit.schemas import Handler, RabbitExchange, RabbitQueue
from propan.log import access_logger
from propan.types import DecodedMessage, DecoratedCallable, Wrapper
from yarl import URL

class RabbitBroker(BrokerUsecase):
    handlers: List[Handler]
    _connection: Optional[aio_pika.RobustConnection]
    _channel: Optional[aio_pika.RobustChannel]

    __max_queue_len: int
    __max_exchange_len: int

    async def __init__(
        self,
        url: Union[str, URL, None] = None,
        host: str = "localhost",
        port: int = 5672,
        login: str = "guest",
        password: str = "guest",
        virtualhost: str = "/",
        ssl: bool = False,
        ssl_options: Optional[aio_pika.abc.SSLOptions] = None,
        ssl_context: Optional[SSLContext] = None,
        timeout: aio_pika.abc.TimeoutType = None,
        client_properties: Optional[FieldTable] = None,
        *,
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        consumers: Optional[int] = None,
    ):
        """
        URL string might be contain ssl parameters e.g.
        `amqps://user:pass@host//?ca_certs=ca.pem&certfile=crt.pem&keyfile=key.pem`

        :param client_properties: add custom client capability.
        :param url:
            RFC3986_ formatted broker address. When :class:`None`
            will be used keyword arguments.
        :param host: hostname of the broker
        :param port: broker port 5672 by default
        :param login: username string. `'guest'` by default.
        :param password: password string. `'guest'` by default.
        :param virtualhost: virtualhost parameter. `'/'` by default
        :param ssl: use SSL for connection. Should be used with addition kwargs.
        :param ssl_options: A dict of values for the SSL connection.
        :param timeout: connection timeout in seconds
        :param ssl_context: ssl.SSLContext instance

        .. _RFC3986: https://goo.gl/MzgYAs
        .. _official Python documentation: https://goo.gl/pty9xA
        """
        ...
    async def connect(
        self,
        url: Union[str, URL, None] = None,
        host: str = "localhost",
        port: int = 5672,
        login: str = "guest",
        password: str = "guest",
        virtualhost: str = "/",
        ssl: bool = False,
        ssl_options: Optional[aio_pika.abc.SSLOptions] = None,
        ssl_context: Optional[SSLContext] = None,
        timeout: aio_pika.abc.TimeoutType = None,
        client_properties: Optional[FieldTable] = None,
    ) -> aio_pika.Connection:
        """
        URL string might be contain ssl parameters e.g.
        `amqps://user:pass@host//?ca_certs=ca.pem&certfile=crt.pem&keyfile=key.pem`

        :param client_properties: add custom client capability.
        :param url:
            RFC3986_ formatted broker address. When :class:`None`
            will be used keyword arguments.
        :param host: hostname of the broker
        :param port: broker port 5672 by default
        :param login: username string. `'guest'` by default.
        :param password: password string. `'guest'` by default.
        :param virtualhost: virtualhost parameter. `'/'` by default
        :param ssl: use SSL for connection. Should be used with addition kwargs.
        :param ssl_options: A dict of values for the SSL connection.
        :param timeout: connection timeout in seconds
        :param ssl_context: ssl.SSLContext instance

        .. _RFC3986: https://goo.gl/MzgYAs
        .. _official Python documentation: https://goo.gl/pty9xA
        """
        ...
    async def publish_message(
        self,
        message: Union[aio_pika.Message, str, Dict[str, Any]],
        queue: Union[RabbitQueue, str] = "",
        exchange: Union[RabbitExchange, str, None] = None,
        *,
        routing_key: str = "",
        mandatory: bool = True,
        immediate: bool = False,
        timeout: aio_pika.abc.TimeoutType = None,
    ) -> Optional[aiormq.abc.ConfirmationFrameType]: ...
    def handle(
        self,
        queue: Union[str, RabbitQueue],
        exchange: Union[str, RabbitExchange, None] = None,
        retry: Union[bool, int] = False,
    ) -> Wrapper:
        """
        retry: Union[bool, int] - at exeption message will returns to queue `int` times or endless if `True`
        """
        ...
    async def __aenter__(self) -> "RabbitBroker": ...
    async def _connect(self, *args: Any, **kwargs: Any) -> aio_pika.Connection: ...
    async def close(self) -> None: ...
    @staticmethod
    async def _decode_message(
        message: aio_pika.IncomingMessage,
    ) -> DecodedMessage: ...
    @staticmethod
    def _process_message(
        func: DecoratedCallable, watcher: Optional[BaseWatcher] = None
    ) -> DecoratedCallable: ...
    def _get_log_context(
        self,
        message: Optional[aio_pika.Message],
        queue: RabbitQueue,
        exchange: Optional[RabbitExchange] = None,
        **kwrags,
    ) -> Dict[str, Any]: ...
    async def _init_channel(self, max_consumers: Optional[int] = None) -> None: ...
    async def _init_handler(
        self, handler: Handler
    ) -> aio_pika.abc.AbstractRobustQueue: ...
    async def _init_queue(
        self, queue: RabbitQueue
    ) -> aio_pika.abc.AbstractRobustQueue: ...
    async def _init_exchange(
        self, exchange: RabbitExchange
    ) -> aio_pika.abc.AbstractRobustExchange: ...
    async def start(self) -> None: ...



"""
Propan has @apply_types decorator to cast incoming function
arguments to type according their type annotation.

If you doesn't create broker as `RabbitBroker(apply_types=False)`,
all broker handlers are wrapped by @apply_types by default.
"""
from propan import PropanApp
from propan.brokers.rabbit import RabbitBroker
from propan.utils import apply_types
from pydantic import BaseModel

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


class SimpleMessage(BaseModel):
    key: int


@app.on_startup
async def publish(broker: RabbitBroker):
    await broker.publish_message("123", queue="test")
    await broker.publish_message(SimpleMessage(key=123).dict(), queue="test2")


@broker.handle("test")
async def base_handler(body: int):
    assert body == 123
    assert _nested_call(body)


@broker.handle("test2")
async def second_handler(body: SimpleMessage):
    assert body.key == 123


@apply_types
def _nested_call(key: str) -> bool:
    return key == "123"



from typing import Any

from fast_depends.library import CustomField
from propan.utils.context import context


class Context(CustomField):
    def __init__(
        self, real_name: str = "", *, cast: bool = False, required: bool = True
    ):
        self.name = real_name
        super().__init__(cast=cast, required=required)

    def use(self, **kwargs):
        name = self.name or self.param_name
        return {**kwargs, self.param_name: resolve_context(name)}


def resolve_context(argument: str) -> Any:
    keys = argument.split(".")

    v = context.context.get(keys[0])
    for i in keys[1:]:
        v = getattr(v, i, None)
        if v is None:
            return v

    return v



import os
import threading
from multiprocessing.context import SpawnProcess
from typing import Any, Optional, Tuple

from propan.cli.supervisors.utils import get_subprocess, set_exit
from propan.log import logger
from propan.types import DecoratedCallable


class BaseReload:
    _process: SpawnProcess
    _target: DecoratedCallable
    _args: Tuple[Any, ...]

    reload_delay: Optional[float]
    should_exit: threading.Event
    pid: int
    reloader_name: str = ""

    def __init__(
        self,
        target: DecoratedCallable,
        args: Tuple[Any, ...],
        reload_delay: Optional[float] = 0.5,
    ) -> None:
        self._target = target
        self._args = args

        self.should_exit = threading.Event()
        self.pid = os.getpid()
        self.reload_delay = reload_delay

        set_exit(lambda *_: self.should_exit.set())

    def run(self) -> None:
        self.startup()
        while not self.should_exit.wait(self.reload_delay):
            if self.should_restart():  # pragma: no branch
                self.restart()
        self.shutdown()

    def startup(self) -> None:
        logger.info(f"Started reloader process [{self.pid}] using {self.reloader_name}")
        self._process = self._start_process()

    def restart(self) -> None:
        self._stop_process()
        logger.info("Process successfully reloaded")
        self._process = self._start_process()

    def shutdown(self) -> None:
        self._stop_process()
        logger.info(f"Stopping reloader process [{self.pid}]")

    def _stop_process(self) -> None:
        self._process.terminate()
        self._process.join()

    def _start_process(self) -> SpawnProcess:
        process = get_subprocess(target=self._target, args=self._args)
        process.start()
        return process

    def should_restart(self) -> bool:
        raise NotImplementedError("Reload strategies should override should_restart()")



from typing import Any, Callable, Dict, TypeVar, Union

from fast_depends.types import DecoratedCallable, P, AnyDict, AnyCallable


__all__ = (
    "DecoratedCallable",
    "P",
    "AnyDict",
    "AnyCallable",
    "DecodedMessage",
    "DecoratedCallableNone"
    "Wrapper"
)

DecoratedCallableNone = TypeVar("DecoratedCallableNone", bound=Callable[..., None])

Wrapper = Callable[[DecoratedCallable], DecoratedCallable]
DecodedMessage = Union[str, Dict[str, Any], bytes]



"""
Depends parameter allows to call function before
natural handler with providing all arguments and context
inside dependency function

Dependecies argument names should be calling same as a handler arguments
Dependency decorated by `apply_types` as a default if hanler was decorated
"""
from propan import Depends, PropanApp
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


async def async_dependency(body, logger):
    logger.info(f"Async calls with: {body}")
    return True


def sync_dependency():
    print("Hello, i am here!")
    return True


@broker.handle("test")
async def base_handler(
    body: dict,
    async_called: bool = Depends(async_dependency),
    sync_called: bool = Depends(sync_dependency),
):
    assert async_called and sync_called



"""
You can use Propan MQBrokers without PropanApp
Just start and stop them whenever you want
"""
from aiohttp import web
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")


@broker.handle("test")
async def base_handler(body):
    print(body)


async def start_broker(app):
    await broker.start()


async def stop_broker(app):
    await broker.close()


async def hello(request):
    return web.Response(text="Hello, world")


app = web.Application()
app.add_routes([web.get("/", hello)])
app.on_startup.append(start_broker)
app.on_cleanup.append(stop_broker)


if __name__ == "__main__":
    web.run_app(app)



import asyncio
import logging
import sys
from pathlib import Path
from typing import Dict, Optional, Union

import typer
from propan.__about__ import __version__
from propan.cli.app import PropanApp
from propan.cli.utils.imports import get_app_path, import_object
from propan.cli.utils.logs import LogLevels, set_log_level
from propan.cli.utils.parser import parse_cli_args
from propan.log import logger

cli = typer.Typer()


def version_callback(version: bool) -> None:
    if version is True:
        import platform

        typer.echo(
            "Running Propan %s with %s %s on %s"
            % (
                __version__,
                platform.python_implementation(),
                platform.python_version(),
                platform.system(),
            )
        )

        raise typer.Exit()


@cli.callback()
def main(
    version: Optional[bool] = typer.Option(
        False,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show current platform, python and propan version",
    )
):
    """
    Generate, run and manage Propan apps to greater development experience
    """


@cli.command()
def create(appname: str) -> None:
    """Create a new Propan project at [APPNAME] directory"""
    from propan.cli.startproject import create

    project = create(Path.cwd() / appname, __version__)
    typer.echo(f"Create Propan project template at: {project}")


@cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def run(
    ctx: typer.Context,
    app: str = typer.Argument(
        ..., help="[python_module:PropanApp] - path to your application"
    ),
    workers: int = typer.Option(
        1, show_default=False, help="Run [workers] applications with process spawning"
    ),
    log_level: LogLevels = typer.Option(
        LogLevels.info, case_sensitive=False, show_default=False, help="[INFO] default"
    ),
    reload: bool = typer.Option(
        False, "--reload", is_flag=True, help="Restart app at directory files changes"
    ),
) -> None:
    """Run [MODULE:APP] Propan application"""
    app, extra = parse_cli_args(app, *ctx.args)

    module, app = get_app_path(app)

    app_dir = module.parent
    sys.path.insert(0, str(app_dir))

    args = (module, app, extra)

    if reload and workers > 1:
        raise ValueError("You can't use reload option with multiprocessing")

    set_log_level(log_level)

    if reload is True:
        from propan.cli.supervisors.watchfiles import WatchReloader

        WatchReloader(target=_run, args=args, reload_dirs=(app_dir,)).run()

    elif workers > 1:
        from propan.cli.supervisors.multiprocess import Multiprocess

        Multiprocess(target=_run, args=(*args, logging.DEBUG), workers=workers).run()

    else:
        _run(*args)


def _run(
    module: Path,
    app: str,
    extra_options: Dict[str, Union[bool, str]],
    log_level: int = logging.INFO,
) -> None:
    try:
        propan_app = import_object(module, app)

        if not isinstance(propan_app, PropanApp):
            raise ValueError(f"{propan_app} is not a PropanApp")

    except (ValueError, FileNotFoundError, AttributeError) as e:
        logger.error(e)
        logger.error("Please, input module like [python_file:propan_app_name]")
        exit()

    else:
        propan_app._command_line_options = extra_options

        if sys.platform not in ("win32", "cygwin", "cli"):
            import uvloop

            if sys.version_info >= (3, 11):
                with asyncio.Runner(loop_factory=uvloop.new_event_loop) as runner:
                    runner.run(propan_app.run(log_level))
                    return

            else:
                uvloop.install()

        asyncio.run(propan_app.run(log_level))



from abc import ABC, abstractmethod
from collections import Counter
from logging import Logger
from types import TracebackType
from typing import Callable, Optional, Type

from propan.utils.functions import call_or_await


class BaseWatcher(ABC):
    max_tries: int

    def __init__(self, max_tries: int = 0, logger: Optional[Logger] = None):
        self.logger = logger
        self.max_tries = max_tries

    @abstractmethod
    def add(self, message_id: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def is_max(self, message_id: str) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def remove(self, message_id: str) -> None:
        raise NotImplementedError()


class FakePushBackWatcher(BaseWatcher):
    def add(self, message_id: str) -> None:
        pass

    def is_max(self, message_id: str) -> bool:
        return False

    def remove(self, message_id: str) -> None:
        pass


class PushBackWatcher(BaseWatcher):
    memory: "Counter[str]"

    def __init__(self, max_tries: int = 3, logger: Optional[Logger] = None):
        super().__init__(logger=logger, max_tries=max_tries)
        self.memory = Counter()

    def add(self, message_id: str) -> None:
        self.memory[message_id] += 1

    def is_max(self, message_id: str) -> bool:
        is_max = self.memory[message_id] > self.max_tries
        if self.logger is not None:
            if is_max:
                self.logger.error(f"Already retried {self.max_tries} times. Skipped.")
            else:
                self.logger.error("Error is occured. Pushing back to queue.")
        return is_max

    def remove(self, message: str) -> None:
        self.memory[message] = 0
        self.memory += Counter()


class WatcherContext:
    def __init__(
        self,
        watcher: BaseWatcher,
        message_id: str,
        on_success: Callable[[], None] = lambda: None,
        on_max: Callable[[], None] = lambda: None,
        on_error: Callable[[], None] = lambda: None,
    ):
        self.watcher = watcher
        self.on_success = on_success
        self.on_max = on_max
        self.on_error = on_error
        self._message_id = message_id

    async def __aenter__(self) -> None:
        self.watcher.add(self._message_id)

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if not exc_type:
            await call_or_await(self.on_success)
            self.watcher.remove(self._message_id)

        elif self.watcher.is_max(self._message_id):
            await call_or_await(self.on_max)
            self.watcher.remove(self._message_id)

        else:
            await call_or_await(self.on_error)



from typing import Any


class Singlethon:
    _instanse = None

    def __new__(cls, *args: Any, **kwargs: Any) -> "Singlethon":
        if cls._instanse is None:
            cls._instanse = super().__new__(cls)
        return cls._instanse

    @classmethod
    def _drop(cls) -> None:
        cls._instanse = None



from .context import Context, ContextRepo, context

from fast_depends import inject as apply_types, Depends

__all__ = (
    "apply_types",
    "context",
    "Context",
    "ContextRepo",
    "Depends",
)



"""
Using propan cli tool allows to pass command-line
options inside your Context dependencies.

Ex:
propan run serve:app --env=.env.dev

You can pass options following ways:
... --env=.env
... -env=.env
... env=.env
"=" is required

Or you can pass a boolean flags
... --use-smth  # passes as use_smth=True
... --no-use-smth  # passes as use_smth=False
"""
from propan import Context, PropanApp
from propan.brokers.rabbit import RabbitBroker
from pydantic import BaseSettings

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


class Settings(BaseSettings):
    ...


@app.on_startup
async def setup_later(env: str, context: Context):
    settings = Settings(_env_file=env)
    context.set_context("settings", settings)



from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import Any, Tuple


def import_object(module: Path, app: str) -> Any:
    spec = spec_from_file_location("mode", f"{module}.py")

    if spec is None:  # pragma: no cover
        raise FileNotFoundError(module)

    mod = module_from_spec(spec)
    loader = spec.loader

    if loader is None:  # pragma: no cover
        raise ValueError(f"{spec} has no loader")

    loader.exec_module(mod)
    obj = getattr(mod, app)

    return obj


def get_app_path(app: str) -> Tuple[Path, str]:
    if ":" not in app:
        raise ValueError(f"{app} is not a PropanApp")

    module, propan_app = app.split(":", 2)

    mod_path = Path.cwd()
    for i in module.split("."):
        mod_path = mod_path / i

    return mod_path, propan_app



import logging
from collections import defaultdict
from enum import Enum
from typing import Union

from propan.log import access_logger, logger


class LogLevels(str, Enum):
    critical = "critical"
    error = "error"
    warning = "warning"
    info = "info"
    debug = "debug"


LOG_LEVELS: "defaultdict[str, int]" = defaultdict(
    lambda: logging.INFO,
    **{
        "critical": logging.CRITICAL,
        "error": logging.ERROR,
        "warning": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.DEBUG,
    },
)


def set_log_level(level: Union[LogLevels, str]) -> None:
    if isinstance(level, LogLevels):
        level = level.value
    log_level = LOG_LEVELS[level]
    logger.setLevel(log_level)
    access_logger.setLevel(log_level)



import multiprocessing
import os
import signal
import sys
from multiprocessing.context import SpawnProcess
from types import FrameType
from typing import Any, Callable, Optional

from propan.types import DecoratedCallableNone

multiprocessing.allow_connection_pickling()
spawn = multiprocessing.get_context("spawn")


HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)


def set_exit(func: Callable[[int, Optional[FrameType]], Any]) -> None:
    for sig in HANDLED_SIGNALS:
        signal.signal(sig, func)


def get_subprocess(target: DecoratedCallableNone, args: Any) -> SpawnProcess:
    stdin_fileno: Optional[int]
    try:
        stdin_fileno = sys.stdin.fileno()
    except OSError:
        stdin_fileno = None

    return spawn.Process(
        target=subprocess_started,
        args=args,
        kwargs={"t": target, "stdin_fileno": stdin_fileno},
    )


def subprocess_started(
    *args: Any,
    t: DecoratedCallableNone,
    stdin_fileno: Optional[int],
) -> None:
    if stdin_fileno is not None:  # pragma: no cover
        sys.stdin = os.fdopen(stdin_fileno)
    t(*args)



"""
All Context dependencies also available from on_startup hook
(Some of them are None at application starting).

You can use it to setup your custom context fields.
"""
from propan import Context, PropanApp
from propan.annotations import ContextRepo
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


@app.on_startup
def setup(context: ContextRepo):
    context.set_context("my_dependency", True)


@broker.handle("test")
async def base_handler(body: dict, my_dependency: bool = Context()):
    assert my_dependency is True



from propan.brokers.rabbit.rabbit_broker import RabbitBroker
from propan.brokers.rabbit.schemas import ExchangeType, RabbitExchange, RabbitQueue

__all__ = (
    "RabbitBroker",
    "RabbitQueue",
    "RabbitExchange",
    "ExchangeType",
)



"""
Using Alias allows you rename context dependencies passing
to your function
"""
from propan import Context, PropanApp
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


@app.on_startup
def setup(rabbit: RabbitBroker = Context("broker")):
    assert rabbit is broker



"""
Connect to localhost RabbitMQ and listen `default` exchange
with `test` routing key

use `propan run basic_usage:app` to start example
"""

from propan import PropanApp
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


@broker.handle("test")
async def base_handler(body):
    print(body)



"""
By default handler drops with rejecting messages due
process exceptions.

If you want to redelivere messages just use
@handler(retry=...) parameter.

For more complex usecases just use the `tenacity` library.
"""

from propan import PropanApp
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


@broker.handle("test", retry=True)
async def base_handler(body):
    """Message will be requeue endlessly at exception"""
    print(body)


@broker.handle("test2", retry=3)
async def another_handler(body):
    """Message will be processed at most 3 times and drops after"""
    print(body)



import asyncio
import json
import logging
from functools import partial, wraps
from typing import Any, Dict, List, Optional, Tuple, Union

import aio_pika
import aiormq
from propan.brokers.model import BrokerUsecase, ContentTypes
from propan.brokers.push_back_watcher import BaseWatcher, WatcherContext
from propan.brokers.rabbit.schemas import Handler, RabbitExchange, RabbitQueue
from propan.types import DecodedMessage, DecoratedCallable, Wrapper
from propan.utils.context.main import log_context


class RabbitBroker(BrokerUsecase):
    handlers: List[Handler]
    _connection: Optional[aio_pika.RobustConnection]
    _channel: Optional[aio_pika.RobustChannel]

    __max_queue_len: int
    __max_exchange_len: int

    def __init__(
        self,
        *args: Any,
        consumers: Optional[int] = None,
        log_fmt: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(*args, log_fmt=log_fmt, **kwargs)
        self._max_consumers = consumers

        self._channel = None
        self.handlers = []

        self.__max_queue_len = 4
        self.__max_exchange_len = 4

    async def __aenter__(self) -> "RabbitBroker":
        await self.connect()
        await self._init_channel()
        return self

    async def _connect(self, *args: Any, **kwargs: Any) -> aio_pika.Connection:
        return await aio_pika.connect_robust(
            *args, **kwargs, loop=asyncio.get_event_loop()
        )

    async def _init_channel(self, max_consumers: Optional[int] = None) -> None:
        if self._channel is None:
            if self._connection is None:
                raise ValueError("RabbitBroker not connected yet")

            max_consumers = max_consumers or self._max_consumers
            self._channel = await self._connection.channel()

            if max_consumers:
                self._log(f"Set max consumers to {max_consumers}", logging.INFO)
                await self._channel.set_qos(prefetch_count=int(max_consumers))

    def handle(
        self,
        queue: Union[str, RabbitQueue],
        exchange: Union[str, RabbitExchange, None] = None,
        retry: Union[bool, int] = False,
    ) -> Wrapper:
        queue, exchange = _validate_exchange_and_queue(queue, exchange)

        if exchange:
            i = len(exchange.name)
            if i > self.__max_exchange_len:  # pragma: no branch
                self.__max_exchange_len = i

        i = len(queue.name)
        if i > self.__max_queue_len:  # pragma: no branch
            self.__max_queue_len = i

        parent = super()

        def wrapper(func: DecoratedCallable) -> Any:
            for handler in self.handlers:
                if handler.exchange == exchange and handler.queue == queue:
                    raise ValueError(
                        f"`{func.__name__}` uses already "
                        f"using `{queue.name}` queue to listen "
                        f"`{exchange.name if exchange else 'default'}` exchange"
                    )

            func = parent._wrap_handler(func, retry, queue=queue, exchange=exchange)
            handler = Handler(callback=func, queue=queue, exchange=exchange)
            self.handlers.append(handler)

        return wrapper

    async def start(self) -> None:
        await super().start()
        await self._init_channel()

        for handler in self.handlers:
            queue = await self._init_handler(handler)

            func = handler.callback

            if self.logger is not None:
                self._get_log_context(None, handler.queue, handler.exchange)
                self.logger.info(f"`{func.__name__}` waiting for messages")

            await queue.consume(func)

    async def publish_message(
        self,
        message: Union[aio_pika.Message, str, Dict[str, Any]],
        queue: Union[RabbitQueue, str] = "",
        exchange: Union[RabbitExchange, str, None] = None,
        *,
        routing_key: str = "",
        **publish_args,
    ) -> Optional[aiormq.abc.ConfirmationFrameType]:
        if self._channel is None:
            raise ValueError("RabbitBroker channel not started yet")

        queue, exchange = _validate_exchange_and_queue(queue, exchange)

        if not isinstance(message, aio_pika.Message):
            if isinstance(message, dict):
                message = aio_pika.Message(
                    json.dumps(message).encode(), content_type=ContentTypes.json.value
                )
            else:
                message = aio_pika.Message(
                    message.encode(), content_type=ContentTypes.text.value
                )

        if exchange is None:
            exchange_obj = self._channel.default_exchange
        else:
            exchange_obj = await self._init_exchange(exchange)

        return await exchange_obj.publish(
            message=message,
            routing_key=routing_key or queue.name,
            **publish_args,
        )

    async def close(self) -> None:
        if self._connection is not None:
            await self._connection.close()

    async def _init_handler(self, handler: Handler) -> aio_pika.abc.AbstractRobustQueue:
        queue = await self._init_queue(handler.queue)
        if handler.exchange is not None:
            exchange = await self._init_exchange(handler.exchange)
            await queue.bind(exchange, handler.queue.routing_key or handler.queue.name)
        return queue

    async def _init_queue(self, queue: RabbitQueue) -> aio_pika.abc.AbstractRobustQueue:
        if queue.declare is True:
            return await self._channel.declare_queue(**queue.dict())
        else:
            return await self._channel.get_queue(queue.name, ensure=False)

    async def _init_exchange(
        self, exchange: RabbitExchange
    ) -> aio_pika.abc.AbstractRobustExchange:
        if exchange.declare is True:
            return await self._channel.declare_exchange(**exchange.dict())
        else:
            return await self._channel.get_exchange(exchange.name, ensure=False)

    def _get_log_context(
        self,
        message: Optional[aio_pika.Message],
        queue: RabbitQueue,
        exchange: Optional[RabbitExchange] = None,
        **kwrags,
    ) -> Dict[str, Any]:
        exchange_name = exchange.name if exchange else "default"
        context = {
            "exchange": exchange_name,
            "queue": queue.name,
            "message_id": message.message_id[:10] if message else "",
        }

        log_context.set(context)
        return context

    @property
    def fmt(self) -> str:
        return self._fmt or (
            "%(asctime)s %(levelname)s - "
            f"%(exchange)-{self.__max_exchange_len}s | "
            f"%(queue)-{self.__max_queue_len}s | "
            f"%(message_id)-10s "
            "- %(message)s"
        )

    @staticmethod
    async def _decode_message(
        message: aio_pika.IncomingMessage,
    ) -> DecodedMessage:
        body = message.body
        if message.content_type is not None:
            if ContentTypes.text.value in message.content_type:
                body = body.decode()
            elif ContentTypes.json.value in message.content_type:  # pragma: no branch
                body = json.loads(body.decode())
        return body

    @staticmethod
    def _process_message(
        func: DecoratedCallable, watcher: Optional[BaseWatcher] = None
    ) -> DecoratedCallable:
        @wraps(func)
        async def wrapper(message: aio_pika.Message):
            if watcher is None:
                context = message.process()
            else:
                context = WatcherContext(
                    watcher,
                    message.message_id,
                    on_success=partial(message.ack),
                    on_error=partial(message.reject, True),
                    on_max=partial(message.reject, False),
                )
            async with context:
                r = await func(message)
                return r

        return wrapper


def _validate_exchange_and_queue(
    queue: Union[str, RabbitQueue, None],
    exchange: Union[str, RabbitExchange, None] = None,
) -> Tuple[RabbitQueue, Optional[RabbitExchange]]:
    if queue is not None:  # pragma: no branch
        if isinstance(queue, str):
            queue = RabbitQueue(name=queue)
        elif not isinstance(queue, RabbitQueue):
            raise ValueError(
                f"Queue '{queue}' should be 'str' | 'RabbitQueue' instance"
            )

    if exchange is not None:  # pragma: no branch
        if isinstance(exchange, str):
            exchange = RabbitExchange(name=exchange)
        elif not isinstance(exchange, RabbitExchange):
            raise ValueError(
                f"Exchange '{exchange}' should be 'str' | 'RabbitExchange' instance"
            )

    return queue, exchange



from pathlib import Path
from typing import Any, Optional, Sequence, Tuple, Union

import watchfiles
from propan.cli.supervisors.basereload import BaseReload
from propan.log import logger
from propan.types import DecoratedCallable


class ExtendedFilter(watchfiles.PythonFilter):
    def __init__(
        self,
        *,
        ignore_paths: Optional[Sequence[Union[str, Path]]] = None,
        extra_extensions: Sequence[str] = (),
    ) -> None:
        super().__init__(ignore_paths=ignore_paths, extra_extensions=extra_extensions)
        self.extensions = self.extensions + (".env", ".yaml")
        self.ignore_dirs = self.ignore_dirs + ("venv", "env", ".ruff_cache", "htmlcov")


class WatchReloader(BaseReload):
    def __init__(
        self,
        target: DecoratedCallable,
        args: Tuple[Any, ...],
        reload_dirs: Sequence[Union[Path, str]],
        reload_delay: Optional[float] = 0.3,
    ) -> None:
        super().__init__(target, args, reload_delay)
        self.reloader_name = "WatchFiles"
        self.watcher = watchfiles.watch(
            *reload_dirs,
            step=int(reload_delay * 1000),
            watch_filter=ExtendedFilter(),
            stop_event=self.should_exit,
            yield_on_timeout=True,
        )

    def should_restart(self) -> bool:
        for changes in self.watcher:  # pragma: no branch
            if changes:  # pragma: no branch
                unique_paths = {Path(c[1]).name for c in changes}
                message = "WatchReloader detected file change in '%s'. Reloading..."
                logger.info(message % tuple(unique_paths))
                return True



from contextvars import ContextVar
from typing import Any, Dict, Optional

from propan.utils.classes import Singlethon

message: ContextVar[Optional[str]] = ContextVar("message", default=None)
log_context: ContextVar[Dict[str, Any]] = ContextVar("message", default={})


class ContextRepo(Singlethon):
    _context: Dict[str, Any] = {}

    def set_context(self, key: str, v: Any) -> None:
        self._context[key] = v

    def remove_context(self, key: str) -> None:
        self._context.pop(key, None)

    def clear(self) -> None:
        self._context = {}

    def __getattr__(self, __name: str) -> Any:
        return self.context.get(__name)

    @property
    def context(self) -> Dict[str, Any]:
        return {**self._context, "context": self, "message": message.get()}


context: ContextRepo = ContextRepo()



from dataclasses import dataclass
from typing import Optional, Sequence

from nats.aio.subscription import Subscription
from nats.js.api import DEFAULT_PREFIX
from propan.types import DecoratedCallable
from pydantic import BaseModel


@dataclass
class Handler:
    callback: DecoratedCallable
    subject: str
    queue: str = ""

    subscription: Optional[Subscription] = None


class JetStream(BaseModel):
    prefix: str = DEFAULT_PREFIX
    domain: Optional[str] = None
    timeout: float = 5

    subjects: Sequence[str] = []



"""
Propan provides the easiest way to publish messages:
just use Broker as context manager and publish whatever you want!
"""
import asyncio

from aio_pika import Message
from propan.brokers.rabbit import RabbitBroker


async def main():
    async with RabbitBroker("amqp://guest:guest@localhost:5672/") as broker:
        await broker.publish_message("Hello, Propan!", queue="test")

        await broker.publish_message(
            {
                "hello": "again!",
            },
            queue="test",
        )

        await broker.publish_message(Message(b"hi"), queue="test")


if __name__ == "__main__":
    asyncio.run(main())



"""
You can use Propan MQBrokers without PropanApp
Just start and stop them whenever you want
"""
from propan.brokers.rabbit import RabbitBroker
from quart import Quart

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = Quart(__name__)


@broker.handle("test")
async def base_handler(body):
    print(body)


@app.before_serving
async def start_broker():
    await broker.start()


@app.after_serving
async def stop_broker():
    await broker.close()


@app.route("/")
async def json():
    return {"hello": "world"}



from functools import partial, wraps
from typing import Callable, Optional

import nats
from nats.aio.msg import Msg
from nats.js.client import JetStreamContext
from propan.brokers.nats.nats_broker import NatsBroker
from propan.brokers.nats.schemas import JetStream
from propan.brokers.push_back_watcher import BaseWatcher, WatcherContext


class NatsJSBroker(NatsBroker):
    _js: Optional[JetStream] = None
    _connection: Optional[JetStreamContext] = None

    def __init__(self, *args, jetstream: JetStream, **kwargs):
        super().__init__(*args, **kwargs)
        self._js = jetstream

    async def _connect(self, *args, **kwargs) -> JetStreamContext:
        nc = await nats.connect(*args, **kwargs)

        stream = await nc.jetstream(
            **self._js.dict(include={"prefix", "domain", "timeout"})
        )

        return stream

    @staticmethod
    def _process_message(
        func: Callable, watcher: Optional[BaseWatcher] = None
    ) -> Callable:
        @wraps(func)
        async def wrapper(message: Msg):
            if watcher is None:
                return await func(message)
            else:
                async with WatcherContext(
                    watcher,
                    message.message_id,
                    on_success=partial(message.ack),
                    on_error=partial(message.nak),
                    on_max=partial(message.term),
                ):
                    await message.in_progress()
                    return await func(message)

        return wrapper



"""
You can use Propan MQBrokers without PropanApp
Just start and stop them whenever you want
"""
from contextlib import asynccontextmanager

from fastapi import FastAPI
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = FastAPI()


@asynccontextmanager
async def lifespan(app: FastAPI):
    await broker.start()
    yield
    await broker.close()


@broker.handle("test")
async def base_handler(body):
    print(body)


@app.get("/")
def read_root():
    return {"Hello": "World"}



import logging
from abc import ABC, abstractmethod
from functools import wraps
from time import perf_counter
from typing import Any, Callable, Dict, List, Optional, Union

from propan.brokers.push_back_watcher import (
    BaseWatcher,
    FakePushBackWatcher,
    PushBackWatcher,
)
from propan.log import access_logger
from propan.types import DecodedMessage, DecoratedCallable, Wrapper
from propan.utils import apply_types
from propan.utils.context import context
from propan.utils.context import message as message_context
from propan.utils.functions import call_or_await


class BrokerUsecase(ABC):
    logger: Optional[logging.Logger]
    log_level: int
    handlers: List[Any]
    _connection: Any
    _fmt: str

    def __init__(
        self,
        *args: Any,
        apply_types: bool = True,
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: str = "%(asctime)s %(levelname)s - %(message)s",
        **kwargs: Any,
    ):
        self.logger = logger
        self.log_level = log_level
        self._fmt = log_fmt

        self._connection = None
        self._is_apply_types = apply_types
        self._connection_args = args
        self._connection_kwargs = kwargs
        self.handlers = []

        context.set_context("logger", logger)
        context.set_context("broker", self)

    async def connect(self, *args: Any, **kwargs: Any) -> Any:
        if self._connection is None:
            _args = args or self._connection_args
            _kwargs = kwargs or self._connection_kwargs
            self._connection = await self._connect(*_args, **_kwargs)
            return self._connection

    @abstractmethod
    async def _connect(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError()

    @abstractmethod
    async def publish_message(self, message: Any, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError()

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def _decode_message(self, message: Any) -> DecodedMessage:
        raise NotImplementedError()

    @abstractmethod
    def _process_message(
        self, func: DecoratedCallable, watcher: Optional[BaseWatcher]
    ) -> Wrapper:
        raise NotImplementedError()

    async def start(self) -> None:
        if self.logger is not None:
            self._init_logger(self.logger)
        await self.connect()

    def _get_log_context(self, **kwargs: Any) -> Dict[str, Any]:
        return {}

    @abstractmethod
    def handle(
        self,
        *broker_args: Any,
        retry: Union[bool, int] = False,
        **broker_kwargs: Any,
    ) -> Callable[[DecoratedCallable], None]:
        raise NotImplementedError()

    @property
    def fmt(self) -> str:
        return self._fmt

    def _init_logger(self, logger: logging.Logger) -> None:
        for handler in logger.handlers:
            formatter = handler.formatter
            if formatter is not None:
                if getattr(formatter, "use_colors", None) is not None:
                    kwargs = {"use_colors": formatter.use_colors}
                else:
                    kwargs = {}
                handler.setFormatter(type(formatter)(self.fmt, **kwargs))

    async def __aenter__(self) -> "BrokerUsecase":
        await self.connect()
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        await self.close()

    def _wrap_handler(
        self, func: DecoratedCallable, retry: Union[bool, int], **broker_args: Any
    ) -> DecoratedCallable:
        if self._is_apply_types is True:
            func = apply_types(func)

        if self.logger is not None:
            func = self._log_execution(self.logger, **broker_args)(func)

        func = self._wrap_decode_message(func)

        func = self._process_message(func, _get_watcher(self.logger, retry))

        func = _set_message_context(func)

        return func

    def _wrap_decode_message(self, func: DecoratedCallable) -> Wrapper:
        @wraps(func)
        async def wrapper(message: Any) -> Any:
            return await func(await self._decode_message(message))

        return wrapper

    def _log_execution(
        self, logger: logging.Logger, **broker_args: Any
    ) -> Callable[[DecoratedCallable], Wrapper]:
        def decor(func: DecoratedCallable) -> Wrapper:
            @wraps(func)
            async def wrapper(*args: Any, **kwargs: Any) -> Any:
                message = message_context.get()

                start = perf_counter()

                self._get_log_context(message=message, **broker_args)
                logger.log(self.log_level, "Received")

                try:
                    r = await call_or_await(func, *args, **kwargs)
                except Exception as e:
                    logger.error(repr(e))
                else:
                    logger.log(
                        self.log_level, f"Processed by {(perf_counter() - start):.4f}"
                    )
                    return r

            return wrapper

        return decor

    def _log(self, message: str, log_level: Optional[int] = None) -> None:
        if self.logger is not None:
            self.logger.log(message, log_level or self.log_level)


def _get_watcher(
    logger: Optional[logging.Logger], try_number: Union[bool, int] = True
) -> Optional[BaseWatcher]:
    watcher: Optional[BaseWatcher]
    if try_number is True:
        watcher = FakePushBackWatcher()
    elif try_number is False:
        watcher = None
    else:
        watcher = PushBackWatcher(logger=logger, max_tries=try_number)
    return watcher


def _set_message_context(func: DecoratedCallable) -> Wrapper:
    @wraps(func)
    async def wrapper(message: Any) -> Any:
        message_context.set(message)
        return await func(message)

    return wrapper



"""
You can use Propan MQBrokers without PropanApp
Just start and stop them whenever you want
"""
from propan.brokers.rabbit import RabbitBroker
from sanic import Sanic
from sanic.response import text

app = Sanic("MyHelloWorldApp")
broker = RabbitBroker("amqp://guest:guest@localhost:5672/")


@broker.handle("test")
async def base_handler(body):
    print(body)


@app.after_server_start
async def start_broker(app, loop):
    await broker.start()


@app.after_server_stop
async def stop_broker(app, loop):
    await broker.close()


@app.get("/")
async def hello_world(request):
    return text("Hello, world.")



from multiprocessing.context import SpawnProcess
from typing import Any, List, Tuple

from propan.cli.supervisors.basereload import BaseReload
from propan.log import logger
from propan.types import DecoratedCallable


class Multiprocess(BaseReload):
    def __init__(
        self, target: DecoratedCallable, args: Tuple[Any, ...], workers: int
    ) -> None:
        super().__init__(target, args, None)

        self.workers = workers
        self.processes: List[SpawnProcess] = []

    def startup(self) -> None:
        logger.info(f"Started parent process [{self.pid}]")

        for _ in range(self.workers):
            process = self._start_process()
            logger.info(f"Started child process [{process.pid}]")
            self.processes.append(process)

    def shutdown(self) -> None:
        for process in self.processes:
            process.terminate()
            logger.info(f"Stopping child process [{process.pid}]")
            process.join()

        logger.info(f"Stopping parent process [{self.pid}]")



"""
You can set specific queue and exchange to listening
with all amqp supported flags
"""
from propan import PropanApp
from propan.brokers.rabbit import (
    ExchangeType,
    RabbitBroker,
    RabbitExchange,
    RabbitQueue,
)

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


@broker.handle(
    queue=RabbitQueue("test"), exchange=RabbitExchange("test", type=ExchangeType.FANOUT)
)
async def base_handler(body: dict):
    print(body)



"""
With Python 3.10+ you can use
typing.Annotated class as a shortcut to Depends and Alias
"""
import logging

from propan import Context, Depends, PropanApp, apply_types
from propan.brokers.rabbit import RabbitBroker
from typing_extensions import Annotated

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)

Logger = Annotated[logging.Logger, Context("broker.logger")]


@apply_types
async def async_dependency(body, logger: Logger):
    logger.info(f"Async calls with: {body}")
    return True


Dependency = Annotated[bool, Depends(async_dependency)]


@broker.handle("test")
async def base_handler(body: dict, async_called: Dependency):
    assert async_called is True



from functools import wraps
from typing import Callable, TypeVar

from fast_depends.injector import run_async as call_or_await
from propan.types import P

T = TypeVar("T")


def to_async(func: Callable[[P], T]) -> Callable[[P], T]:
    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return await call_or_await(func, *args, **kwargs)

    return wrapper



"""
You can use Propan MQBrokers without PropanApp
Just start and stop them whenever you want
"""
from blacksheep import Application
from propan.brokers.rabbit import RabbitBroker

app = Application()


broker = RabbitBroker("amqp://guest:guest@localhost:5672/")


@broker.handle("test")
async def base_handler(body):
    print(body)


@app.on_start
async def start_broker(application: Application) -> None:
    await broker.start()


@app.on_stop
async def stop_broker(application: Application) -> None:
    await broker.close()


@app.route("/")
async def home():
    return "Hello, World!"



from enum import Enum
from typing import Any, Optional

from pydantic import BaseModel, Field


class ContentTypes(str, Enum):
    text = "text/plain"
    json = "application/json"


class NameRequired(BaseModel):
    name: Optional[str] = Field(...)

    def __init__(self, name: str, **kwargs: Any):
        super().__init__(name=name, **kwargs)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NameRequired) and self.name == other.name


class Queue(NameRequired):
    name: Optional[str] = Field(...)

    def __init__(self, name: str, **kwargs: Any):
        super().__init__(name=name, **kwargs)



"""
Also you can use sync/async functions as
application lifecycle hooks

All `on_startup` hooks runs before broker has been started
All `on_shutdown` hooks runs after broker has been stopped
"""

from propan import PropanApp
from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)


@app.on_startup
def setup():
    print("Application starting up...")


@app.on_startup
async def setup_later():
    print("Application still starting up...")


@app.on_shutdown
async def shutdown():
    print("Application shutting down...")


@broker.handle("test")
async def base_handler(body: dict):
    print(body)



import logging
import sys
from collections import defaultdict
from typing import Callable, Optional

import click
from propan.utils.context.main import log_context


class ColourizedFormatter(logging.Formatter):
    level_name_colors: "defaultdict[str, Callable[[str], str]]" = defaultdict(
        lambda: str,
        **{
            str(logging.DEBUG): lambda level_name: click.style(
                str(level_name), fg="cyan"
            ),
            str(logging.INFO): lambda level_name: click.style(
                str(level_name), fg="green"
            ),
            str(logging.WARNING): lambda level_name: click.style(
                str(level_name), fg="yellow"
            ),
            str(logging.ERROR): lambda level_name: click.style(
                str(level_name), fg="red"
            ),
            str(logging.CRITICAL): lambda level_name: click.style(
                str(level_name), fg="bright_red"
            ),
        },
    )

    def __init__(
        self,
        fmt: Optional[str] = None,
        datefmt: Optional[str] = None,
        style: str = "%",
        use_colors: Optional[bool] = None,
    ):
        """
        Initialize the formatter with specified format strings.

        Initialize the formatter either with the specified format string, or a
        default as described above. Allow for specialized date formatting with
        the optional datefmt argument. If datefmt is omitted, you get an
        ISO8601-like (or RFC 3339-like) format.

        Use a style parameter of '%', '{' or '$' to specify that you want to
        use one of %-formatting, :meth:`str.format` (``{}``) formatting or
        :class:`string.Template` formatting in your format string.
        """
        if use_colors in (True, False):
            self.use_colors = use_colors
        else:
            self.use_colors = sys.stdout.isatty()
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)

    def color_level_name(self, level_name: str, level_no: int) -> str:
        return self.level_name_colors[str(level_no)](level_name)

    def formatMessage(self, record: logging.LogRecord) -> str:
        levelname = expand_log_field(record.levelname, 8)
        if self.use_colors is True:  # pragma: no cover
            levelname = self.color_level_name(levelname, record.levelno)
        record.__dict__["levelname"] = levelname
        return super().formatMessage(record)


class DefaultFormatter(ColourizedFormatter):
    pass


class AccessFormatter(ColourizedFormatter):
    def formatMessage(self, record: logging.LogRecord) -> str:
        context = log_context.get()
        if context:  # pragma: no cover
            record.__dict__.update(context)
        return super().formatMessage(record)


def expand_log_field(field: str, symbols: int) -> str:
    return field + (" " * (symbols - len(field)))



import logging
import ssl
from typing import Any, Callable, Dict, List, Optional, Union

from nats.aio.client import (
    DEFAULT_CONNECT_TIMEOUT,
    DEFAULT_DRAIN_TIMEOUT,
    DEFAULT_INBOX_PREFIX,
    DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
    DEFAULT_MAX_OUTSTANDING_PINGS,
    DEFAULT_MAX_RECONNECT_ATTEMPTS,
    DEFAULT_PENDING_SIZE,
    DEFAULT_PING_INTERVAL,
    DEFAULT_RECONNECT_TIME_WAIT,
    Callback,
    Client,
    Credentials,
    ErrorCallback,
    JWTCallback,
    SignatureCallback,
)
from nats.aio.msg import Msg
from propan.brokers.model import BrokerUsecase
from propan.brokers.nats.schemas import Handler
from propan.brokers.push_back_watcher import BaseWatcher
from propan.log import access_logger
from propan.types import DecoratedCallable

class NatsBroker(BrokerUsecase):
    logger: logging.Logger
    handlers: List[Handler] = []
    _connection: Optional[Client] = None

    def __init__(
        self,
        servers: Union[str, List[str]] = ["nats://localhost:4222"],  # noqa: B006
        error_cb: Optional[ErrorCallback] = None,
        disconnected_cb: Optional[Callback] = None,
        closed_cb: Optional[Callback] = None,
        discovered_server_cb: Optional[Callback] = None,
        reconnected_cb: Optional[Callback] = None,
        name: Optional[str] = None,
        pedantic: bool = False,
        verbose: bool = False,
        allow_reconnect: bool = True,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        reconnect_time_wait: int = DEFAULT_RECONNECT_TIME_WAIT,
        max_reconnect_attempts: int = DEFAULT_MAX_RECONNECT_ATTEMPTS,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        max_outstanding_pings: int = DEFAULT_MAX_OUTSTANDING_PINGS,
        dont_randomize: bool = False,
        flusher_queue_size: int = DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
        no_echo: bool = False,
        tls: Optional[ssl.SSLContext] = None,
        tls_hostname: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        drain_timeout: int = DEFAULT_DRAIN_TIMEOUT,
        signature_cb: Optional[SignatureCallback] = None,
        user_jwt_cb: Optional[JWTCallback] = None,
        user_credentials: Optional[Credentials] = None,
        nkeys_seed: Optional[str] = None,
        inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX,
        pending_size: int = DEFAULT_PENDING_SIZE,
        flush_timeout: Optional[float] = None,
        *,
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        consumers: Optional[int] = None,
    ) -> None: ...
    async def connect(
        self,
        servers: Union[str, List[str]] = ["nats://localhost:4222"],  # noqa: B006
        error_cb: Optional[ErrorCallback] = None,
        disconnected_cb: Optional[Callback] = None,
        closed_cb: Optional[Callback] = None,
        discovered_server_cb: Optional[Callback] = None,
        reconnected_cb: Optional[Callback] = None,
        name: Optional[str] = None,
        pedantic: bool = False,
        verbose: bool = False,
        allow_reconnect: bool = True,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        reconnect_time_wait: int = DEFAULT_RECONNECT_TIME_WAIT,
        max_reconnect_attempts: int = DEFAULT_MAX_RECONNECT_ATTEMPTS,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        max_outstanding_pings: int = DEFAULT_MAX_OUTSTANDING_PINGS,
        dont_randomize: bool = False,
        flusher_queue_size: int = DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
        no_echo: bool = False,
        tls: Optional[ssl.SSLContext] = None,
        tls_hostname: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        drain_timeout: int = DEFAULT_DRAIN_TIMEOUT,
        signature_cb: Optional[SignatureCallback] = None,
        user_jwt_cb: Optional[JWTCallback] = None,
        user_credentials: Optional[Credentials] = None,
        nkeys_seed: Optional[str] = None,
        inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX,
        pending_size: int = DEFAULT_PENDING_SIZE,
        flush_timeout: Optional[float] = None,
    ) -> Client: ...
    async def publish_message(
        self,
        message: Union[str, Dict[str, Any]],
        subject: str,
        reply: str = "",
        headers: Optional[Dict[str, str]] = None,
    ) -> None: ...
    def handle(
        self, subject: str, queue: str = "", retry: Union[bool, int] = False
    ) -> Callable[[DecoratedCallable], None]: ...
    async def __aenter__(self) -> "NatsBroker": ...
    async def _connect(self, *args: Any, **kwargs: Any) -> Client: ...
    async def close(self) -> None: ...
    def _get_log_context(
        self, message: Optional[Msg], subject: str, queue: str = "", **kwargs
    ) -> Dict[str, Any]: ...
    @staticmethod
    async def _decode_message(message: Msg) -> Union[str, dict]: ...
    @staticmethod
    def _process_message(
        func: DecoratedCallable, watcher: Optional[BaseWatcher] = None
    ) -> Callable[[Msg], Any]: ...



from typing import Any, Dict, Optional, Union

from aio_pika.abc import ExchangeType, TimeoutType
from propan.brokers.model.schemas import NameRequired, Queue
from propan.types import DecoratedCallable
from pydantic import BaseModel, Field

__all__ = (
    "RabbitQueue",
    "RabbitExchange",
    "Handler",
    "ExchangeType",
)


class RabbitQueue(Queue):
    durable: bool = False
    exclusive: bool = False
    passive: bool = False
    auto_delete: bool = False
    arguments: Optional[Dict[str, Any]] = None
    timeout: TimeoutType = None
    robust: bool = True

    declare: bool = Field(default=True, exclude=True)
    routing_key: str = Field(default="", exclude=True)


class RabbitExchange(NameRequired):
    type: Union[ExchangeType, str] = ExchangeType.DIRECT
    durable: bool = False
    auto_delete: bool = False
    internal: bool = False
    passive: bool = False
    arguments: Optional[Dict[str, Any]] = None
    timeout: TimeoutType = None
    robust: bool = True

    declare: bool = Field(default=True, exclude=True)


class Handler(BaseModel):
    callback: DecoratedCallable
    queue: RabbitQueue
    exchange: Optional[RabbitExchange] = None



from pathlib import Path
from typing import Union


def create(project_dir: Path, version: str) -> Path:
    project_dir = _create_project_dir(project_dir, version)
    app_dir = _create_app_dir(project_dir / "app")
    _create_config_dir(app_dir / "config")
    _create_core_dir(app_dir / "core")
    _create_apps_dir(app_dir / "apps")
    return project_dir


def _create_project_dir(dirname: Path, version: str) -> Path:
    project_dir = _touch_dir(dirname)

    _write_file(project_dir / "README.md")

    _write_file(
        project_dir / "Dockerfile",
        "FROM python:3.11.3-slim",
        "",
        "RUN useradd -ms /bin/bash user",
        "",
        "USER user",
        "WORKDIR /home/user",
        "",
        "COPY requirements.txt requirements.txt",
        "",
        "RUN pip install -r requirements.txt",
        "",
        "COPY ./app ./app",
        "",
        'ENTRYPOINT [ "python", "-m", "propan", "run", "app.serve:app" ]',
    )

    _write_file(
        project_dir / "docker-compose.yaml",
        'version: "3"',
        "",
        "services:",
        "  rabbit:",
        "    image: rabbitmq",
        "    environment:",
        "      RABBITMQ_DEFAULT_USER: guest",
        "      RABBITMQ_DEFAULT_PASS: guest",
        "    ports:",
        "      - 5672:5672",
        "",
        "  app:",
        "    build: .",
        "    environment:",
        "      APP_RABBIT__URL: amqp://guest:guest@rabbit:5672/",
        "    volumes:",
        "      - ./app:/home/user/app:ro",
        "    depends_on:",
        "      - rabbit",
    )

    _write_file(
        project_dir / "requirements.txt",
        f"propan[async_rabbit]=={version}",
        "pydantic[dotenv]",
    )

    _write_file(
        project_dir / ".gitignore",
        "*.env*",
        "__pycache__",
        "env/",
        "venv/",
    )

    return project_dir


def _create_app_dir(app: Path) -> Path:
    app_dir = _touch_dir(app)

    _write_file(
        app_dir / "serve.py",
        "import logging",
        "from typing import Optional",
        "",
        "from propan import PropanApp",
        "from propan.utils import Context",
        "from propan.brokers.rabbit import RabbitBroker",
        "",
        "from core import broker",
        "from config import init_settings",
        "",
        "from apps import *  # import to register handlers  # NOQA",
        "",
        "",
        "app = PropanApp(broker)",
        "",
        "",
        "@app.on_startup",
        "async def init_app(broker: RabbitBroker, env: Optional[str], context: Context):",
        "    settings = init_settings(env)",
        '    context.set_context("settings", settings)',
        "",
        "    logger_level = logging.DEBUG if settings.debug else logging.INFO",
        "    app.logger.setLevel(logger_level)",
        "    broker.logger.setLevel(logger_level)",
        "",
        "    await broker.connect(url=settings.rabbit.url)",
        "",
        "",
        'if __name__ == "__main__":',
        "    app.run()",
    )

    return app_dir


def _create_config_dir(config: Path) -> Path:
    config_dir = _touch_dir(config)

    _write_file(
        config_dir / ".env",
        "APP_DEBUG=True",
        "APP_RABBIT__URL=amqp://guest:guest@localhost:5672/",
    )

    _write_file(
        config_dir / "settings.py",
        "from pathlib import Path",
        "from typing import Optional",
        "",
        "from pydantic import BaseSettings, BaseModel, Field",
        "",
        "",
        "CONFIG_DIR = Path(__file__).resolve().parent",
        "BASE_DIR = CONFIG_DIR.parent",
        "",
        "",
        "class RabbitSettings(BaseModel):",
        "    url: str = Field(...)",
        "",
        "",
        "class Settings(BaseSettings):",
        "    debug: bool = Field(True)",
        "    rabbit: RabbitSettings = Field(default_factory=RabbitSettings)",
        "    base_dir: Path = BASE_DIR",
        "",
        "    class Config:",
        "        env_prefix = 'APP_'",
        "        env_file_encoding = 'utf-8'",
        "        env_nested_delimiter = '__'",
        "",
        "",
        "def init_settings(env_file: Optional[str]) -> Settings:",
        '    env_file = CONFIG_DIR / (env_file or ".env")',
        "    if env_file.exists():",
        "        settings = Settings(_env_file=env_file)",
        "    else:",
        "        settings = Settings()",
        "    return settings",
    )

    _write_file(
        config_dir / "__init__.py",
        "from .settings import Settings, init_settings",
    )

    return config_dir


def _create_core_dir(core: Path) -> Path:
    core_dir = _touch_dir(core)

    _write_file(
        core_dir / "__init__.py",
        "from .dependencies import broker",
    )

    _write_file(
        core_dir / "dependencies.py",
        "from propan.brokers.rabbit import RabbitBroker",
        "",
        "broker = RabbitBroker()",
    )

    return core_dir


def _create_apps_dir(apps: Path) -> Path:
    apps_dir = _touch_dir(apps)

    _write_file(
        apps_dir / "__init__.py",
        "from .handlers import base_handler",
    )

    _write_file(
        apps_dir / "handlers.py",
        "from core import broker",
        "",
        "from propan.brokers.rabbit import RabbitQueue, RabbitExchange",
        "",
        "",
        '@broker.handle(queue=RabbitQueue("test"),',
        '               exchange=RabbitExchange("test"))',
        "async def base_handler(body: dict, logger):",
        "    logger.info(body)",
    )

    return apps_dir


def _touch_dir(dir: Union[Path, str]) -> Path:
    if isinstance(dir, str) is True:
        dir = Path(dir).resolve()

    if dir.exists() is False:
        dir.mkdir()
    return dir


def _write_file(path: Path, *content: str) -> None:
    path.touch()
    if content:
        path.write_text("\n".join(content))



"""
Propan use dependencies management policy close to `pytest fixtures`.
You can specify in functions argument parameters which dependencies
you would to use. And framework passes them from the global Context object.

Default context fields are: app, broker, context (itself), logger and message.
If you call not existed field, raises "pydantic.error_wrappers.ValidationError" value.
"""
from propan import Context, PropanApp
from propan.annotations import App, ContextRepo, Logger
from propan.annotations import RabbitBroker as Rabbit
from propan.annotations import RabbitMessage
from propan.brokers.rabbit import RabbitBroker

rabbit_broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(rabbit_broker)


@rabbit_broker.handle("test")
async def base_handler(
    body: dict,
    app: App,
    broker: Rabbit,
    context: ContextRepo,
    logger: Logger,
    message: RabbitMessage,
    not_found_field=Context(required=False),
):
    assert not_found_field is None
    assert broker is rabbit_broker
